import json
import unittest
from tempfile import TemporaryDirectory
from unittest import mock
import numpy as np

from twined import exceptions
from twined.utils import TwinedEncoder, load_json
from .base import VALID_SCHEMA_TWINE, BaseTestCase


class TestUtils(BaseTestCase):
    """Testing operation of the Twine class"""

    def test_load_json_with_file_like(self):
        """Ensures that json can be loaded from a file-like object"""
        with TemporaryDirectory() as tmp_dir:
            with open(self._write_json_string_to_file(VALID_SCHEMA_TWINE, tmp_dir), "r") as file_like:
                data = load_json(file_like)
                for key in data.keys():
                    self.assertIn(key, ("configuration_values_schema", "input_values_schema", "output_values_schema"))

    def test_load_json_with_object(self):
        """Ensures if load_json is called on an already loaded object, it'll pass-through successfully"""
        already_loaded_data = {"a": 1, "b": 2}
        data = load_json(already_loaded_data)
        for key in data.keys():
            self.assertIn(key, ("a", "b"))

    def test_load_json_with_disallowed_kind(self):
        """Ensures that when attempting to load json with a kind which is diallowed, the correct exception is raised"""
        custom_allowed_kinds = ("file-like", "filename", "object")  # Removed  "string"
        with self.assertRaises(exceptions.InvalidSourceKindException):
            load_json("{}", allowed_kinds=custom_allowed_kinds)

    def test_encoder_without_numpy(self):
        """Ensures that the json encoder can work without numpy being installed"""
        some_json = {"a": np.array([0, 1])}
        with mock.patch("twined.utils.encoders._numpy_spec", new=None):
            with self.assertRaises(TypeError) as e:
                json.dumps(some_json, cls=TwinedEncoder)

        # Very annoying behaviour change between python 3.6 and 3.8
        py38 = "Object of type 'ndarray' is not JSON serializable" in e.exception.args[0]
        py36 = "Object of type ndarray is not JSON serializable" in e.exception.args[0]
        self.assertTrue(py36 or py38)

    def test_encoder_with_numpy(self):
        """Ensures that the json encoder can work with numpy installed"""
        some_json = {"a": np.array([0, 1])}
        json.dumps(some_json, cls=TwinedEncoder)


if __name__ == "__main__":
    unittest.main()



from twined import Twine, exceptions
from .base import BaseTestCase


class TestMonitorMessageTwine(BaseTestCase):
    STRAND_WITH_MONITOR_MESSAGE_SCHEMA = """
        {
            "monitor_message_schema": {
                "type": "object",
                "properties": {
                    "my_property": {
                        "type": "number"
                    }
                },
                "required": ["my_property"]
            }
        }
    """

    def test_validate_monitor_message_raises_error_if_monitor_message_schema_not_met(self):
        """Test that an error is raised if an invalid monitor update is validated."""
        twine = Twine(source=self.STRAND_WITH_MONITOR_MESSAGE_SCHEMA)

        with self.assertRaises(exceptions.InvalidValuesContents):
            twine.validate_monitor_message([])

    def test_validate_monitor_message_with_valid_monitor_update(self):
        """Test that a valid monitor update validates successfully."""
        twine = Twine(source=self.STRAND_WITH_MONITOR_MESSAGE_SCHEMA)
        twine.validate_monitor_message({"my_property": 3.7})



import os
import unittest
from unittest import mock

from twined import Twine, exceptions
from .base import VALID_SCHEMA_TWINE, BaseTestCase


class TestCredentialsTwine(BaseTestCase):
    """Tests related to the twine itself - ensuring that valid and invalid `credentials` entries in a twine file work
    as expected.
    """

    def test_fails_on_no_name(self):
        """Ensures InvalidTwine exceptions are raised when instantiating twines with a missing `name` field in a
        credential.
        """
        invalid_credentials_no_name_twine = """
            {
                "credentials": [
                    {
                        "purpose": "credentials without a name should be invalid"
                    }
                ]
            }
        """

        with self.assertRaises(exceptions.InvalidTwine):
            Twine(source=invalid_credentials_no_name_twine)

    def test_fails_on_lowercase_name(self):
        """Ensures InvalidTwine exceptions are raised when instantiating twines with lowercase letters in the `name`
        field.
        """
        invalid_credentials_lowercase_name_twine = """
            {
                "credentials": [
                    {
                        "name": "my_secrets_should_be_uppercase",
                        "purpose": "Token for accessing a 3rd party API service"
                    }
                ]
            }
        """

        with self.assertRaises(exceptions.InvalidTwine):
            Twine(source=invalid_credentials_lowercase_name_twine)

    def test_fails_on_dict(self):
        """Ensures InvalidTwine exceptions are raised when instantiating twines with invalid `credentials` entries
        (given as a dict, not an array).
        """
        invalid_credentials_dict_not_array_twine = """
            {
                "credentials": {
                    "name": "MY_API_SECRET_KEY",
                    "purpose": "Token for accessing a 3rd party API service"
                }
            }
        """

        with self.assertRaises(exceptions.InvalidTwine):
            Twine(source=invalid_credentials_dict_not_array_twine)

    def test_fails_on_name_whitespace(self):
        """Test that a credential with spaces in its name causes an error to be raised when validated."""
        invalid_credentials_space_in_name_twine = """
            {
                "credentials": [
                    {
                        "name": "MY NAME SHOULD NOT HAVE WHITESPACE",
                        "purpose": "Token for accessing a 3rd party API service"
                    }
                ]
            }
        """

        with self.assertRaises(exceptions.InvalidTwine):
            Twine(source=invalid_credentials_space_in_name_twine)


class TestCredentialsValidation(BaseTestCase):
    """Tests related to whether validation of children occurs successfully (given a valid twine)"""

    VALID_CREDENTIALS_TWINE = """
        {
            "credentials": [
                {
                    "name": "SECRET_THE_FIRST",
                    "purpose": "Token for accessing a 3rd party API service"
                },
                {
                    "name": "SECRET_THE_SECOND",
                    "purpose": "Token for accessing a 3rd party API service"
                },
                {
                    "name": "SECRET_THE_THIRD"
                }
            ]
        }
    """

    def test_no_credentials(self):
        """Test that a twine with no credentials will validate straightforwardly"""
        twine = Twine(source=VALID_SCHEMA_TWINE)
        twine.validate_credentials()

    def test_missing_credentials(self):
        """Test that a twine with credentials will not validate where they are missing from the environment"""
        twine = Twine(source=self.VALID_CREDENTIALS_TWINE)
        with self.assertRaises(exceptions.CredentialNotFound):
            twine.validate_credentials()

    def test_credentials(self):
        """Test that the environment will override a default value for a credential."""
        twine = Twine(source=self.VALID_CREDENTIALS_TWINE)
        with mock.patch.dict(
            os.environ,
            {"SECRET_THE_FIRST": "a value", "SECRET_THE_SECOND": "another value", "SECRET_THE_THIRD": "value"},
        ):
            twine.validate_credentials()
            self.assertEqual(os.environ["SECRET_THE_THIRD"], "value")


if __name__ == "__main__":
    unittest.main()



import unittest

from twined import Twine, exceptions
from .base import BaseTestCase


class TestChildrenTwine(BaseTestCase):
    """Tests ensuring that valid and invalid `children` entries in a twine file work as expected."""

    def test_invalid_children_dict_not_array(self):
        """Ensure that `InvalidTwine` exceptions are raised when instantiating twines where `children` entry is
        incorrectly specified as a dict, not an array.
        """
        with self.assertRaises(exceptions.InvalidTwine):
            Twine(source="""{"children": {}}""")

    def test_invalid_children_no_key(self):
        """Ensure that `InvalidTwine` exceptions are raised when instantiating twines where a child is specified without
        the required `key` field.
        """
        source = """
            {
                "children": [{"purpose": "The purpose.", "notes": "Here are some notes."}]
            }
        """

        with self.assertRaises(exceptions.InvalidTwine):
            Twine(source=source)

    def test_valid_children(self):
        """Ensures that a twine with one child can be instantiated correctly."""
        source = """
            {
                "children": [{"key": "gis", "purpose": "The purpose.", "notes": "Some notes."}]
            }
        """
        self.assertEqual(len(Twine(source=source).children), 1)

    def test_empty_children(self):
        """Ensures that a twine file will validate with an empty list object as children"""
        twine = Twine(source="""{"children": []}""")
        self.assertEqual(len(twine.children), 0)


class TestChildrenValidation(BaseTestCase):
    """Tests related to whether validation of children occurs successfully (given a valid twine)"""

    VALID_TWINE_WITH_CHILDREN = """
        {
            "children": [{"key": "gis", "purpose": "The purpose", "notes": "Some notes."}]
        }
    """

    VALID_CHILD_VALUE = """
        [
            {
                "key": "gis",
                "id": "some-id",
                "backend": {
                    "name": "GCPPubSubBackend",
                    "project_name": "my-project"
                }
            }
        ]
    """

    def test_no_children(self):
        """Test that a twine with no children will validate on an empty children input"""
        Twine().validate_children(source=[])

    def test_missing_children(self):
        """Test that a twine with children will not validate on an empty children input."""
        with self.assertRaises(exceptions.InvalidValuesContents):
            Twine(source=self.VALID_TWINE_WITH_CHILDREN).validate_children(source=[])

    def test_extra_children(self):
        """Test that a twine with no children will not validate a non-empty children input."""
        with self.assertRaises(exceptions.InvalidValuesContents):
            Twine().validate_children(source=self.VALID_CHILD_VALUE)

    def test_backend_cannot_be_empty(self):
        """Test that the backend field of a child cannot be empty."""
        single_child_missing_backend = """[{"key": "gis", "id": "some-id", "backend": {}}]"""

        with self.assertRaises(exceptions.InvalidValuesContents):
            Twine().validate_children(source=single_child_missing_backend)

    def test_extra_key_validation_on_empty_twine(self):
        """Test that children with extra data will not raise a validation error on an empty twine."""
        children_values_with_extra_data = """
            [
                {"key": "gis", "id": "id", "uri_env_name": "VAR_NAME", "an_extra_key": "not a problem if present"},
                {"key": "some_weird_other_child", "id": "some-other-id", "uri_env_name": "SOME_ENV_VAR_NAME"}
            ]
        """

        with self.assertRaises(exceptions.InvalidValuesContents):
            Twine().validate_children(source=children_values_with_extra_data)

    def test_extra_key_validation_on_valid_twine(self):
        """Test that children with extra data will not raise a validation error on a non-empty valid twine.
        # TODO review this behaviour - possibly should raise an error but allow for a user specified extra_data property
        """
        single_child_with_extra_data = """
            [
                {
                    "key": "gis",
                    "id": "some-id",
                    "backend": {
                        "name": "GCPPubSubBackend",
                        "project_name": "my-project"
                    },
                    "some_extra_property": "should not be a problem if present"
                }
            ]
        """

        twine = Twine(source=self.VALID_TWINE_WITH_CHILDREN)
        twine.validate_children(source=single_child_with_extra_data)

    def test_invalid_env_name(self):
        """Test that a child uri env name not in ALL_CAPS_SNAKE_CASE doesn't validate"""
        child_with_invalid_environment_variable_name = """
            [
                {
                    "key": "gis",
                    "id": "some-id",
                    "uri_env_name": "an environment variable not in CAPS_CASE is invalid per the credentials spec"
                }
            ]
        """

        with self.assertRaises(exceptions.InvalidValuesContents):
            Twine().validate_children(source=child_with_invalid_environment_variable_name)

    def test_invalid_json(self):
        """Tests that a children entry with invalid json will raise an error"""
        with self.assertRaises(exceptions.InvalidValuesJson):
            Twine(source=self.VALID_TWINE_WITH_CHILDREN).validate_children(source="[")

    def test_valid(self):
        """Test that a valid twine will validate valid children
        Valiantly and Validly validating validity since 1983.
        To those reading this, know that YOU'RE valid.
        """
        twine = Twine(source=self.VALID_TWINE_WITH_CHILDREN)
        twine.validate_children(source=self.VALID_CHILD_VALUE)


if __name__ == "__main__":
    unittest.main()



import os
import unittest
from tempfile import TemporaryDirectory

from twined import Twine, exceptions
from .base import VALID_SCHEMA_TWINE, BaseTestCase


class TestSchemaStrands(BaseTestCase):
    """Testing operation of the Twine class for validation of data using strands which contain schema"""

    VALID_CONFIGURATION_VALUE = """{"n_iterations": 1}"""

    def test_invalid_strand(self):
        """Ensures that an incorrect strand name would lead to the correct exception
        Note: This tests an internal method. The current API doesn't allow this error to emerge but tthis check allows
        us to extend to a generic method
        """
        twine = Twine(source=VALID_SCHEMA_TWINE)
        data = twine._load_json("configuration", source=self.VALID_CONFIGURATION_VALUE)
        with self.assertRaises(exceptions.UnknownStrand):
            twine._validate_against_schema("not_a_strand_name", data)

    def test_missing_values_files(self):
        """Ensures that if you try to read values from missing files, the right exceptions get raised"""
        twine = Twine(source=VALID_SCHEMA_TWINE)
        values_file = os.path.join(self.path, "not_a_file.json")

        with self.assertRaises(exceptions.ConfigurationValuesFileNotFound):
            twine.validate_configuration_values(source=values_file)

        with self.assertRaises(exceptions.InputValuesFileNotFound):
            twine.validate_input_values(source=values_file)

        with self.assertRaises(exceptions.OutputValuesFileNotFound):
            twine.validate_output_values(source=values_file)

    def test_no_values(self):
        """Ensures that giving no data source raises an invalidJson error"""
        with self.assertRaises(exceptions.InvalidValuesJson):
            Twine(source=VALID_SCHEMA_TWINE).validate_configuration_values(source=None)

    def test_empty_values(self):
        """Ensures that appropriate errors are generated for invalid values"""
        with self.assertRaises(exceptions.InvalidValuesJson):
            Twine(source=VALID_SCHEMA_TWINE).validate_configuration_values(source="")

    def test_strand_not_found(self):
        """Ensures that if a twine doesn't have a strand, you can't validate against it"""
        valid_no_output_schema_twine = """
           {
                "configuration_values_schema": {
                    "$schema": "https://json-schema.org/draft/2020-12/schema",
                    "title": "The example configuration form",
                    "description": "The configuration strand of an example twine",
                    "type": "object",
                    "properties": {
                        "n_iterations": {
                            "description": "An example of an integer configuration variable, called 'n_iterations'.",
                            "type": "integer",
                            "minimum": 1,
                            "maximum": 10,
                            "default": 5
                        }
                    }
                }
            }
        """

        with self.assertRaises(exceptions.StrandNotFound):
            Twine(source=valid_no_output_schema_twine).validate_output_values(source="{}")

    def test_incorrect_values(self):
        """Ensures that appropriate errors are generated for invalid values"""
        incorrect_configuration_value = """{"n_iterations": "should not be a string, this field requires an integer"}"""
        with self.assertRaises(exceptions.InvalidValuesContents):
            Twine(source=VALID_SCHEMA_TWINE).validate_configuration_values(source=incorrect_configuration_value)

    def test_missing_not_required_values(self):
        """Ensures that appropriate errors are generated for missing values"""
        Twine(source=VALID_SCHEMA_TWINE).validate_output_values(source="{}")

    def test_missing_required_values(self):
        """Ensures that appropriate errors are generated for missing values"""
        with self.assertRaises(exceptions.InvalidValuesContents):
            Twine(source=VALID_SCHEMA_TWINE).validate_input_values(source="{}")

    def test_valid_values_files(self):
        """Ensures that values can be read and validated correctly from files on disk"""
        twine = Twine(source=VALID_SCHEMA_TWINE)

        with TemporaryDirectory() as tmp_dir:
            valid_configuration_file = self._write_json_string_to_file(self.VALID_CONFIGURATION_VALUE, tmp_dir)
            twine.validate_configuration_values(source=valid_configuration_file)
            twine.validate_input_values(source="""{"height": 40}""")
            twine.validate_output_values(source="""{"width": 36}""")

    def test_valid_values_json(self):
        """Ensures that values can be read and validated correctly from a json string"""
        Twine(source=VALID_SCHEMA_TWINE).validate_configuration_values(source=self.VALID_CONFIGURATION_VALUE)

    def test_valid_with_extra_values(self):
        """Ensures that extra values get ignored"""
        configuration_valid_with_extra_field = """
            {
                "n_iterations": 1,
                "another_field": "may or may not be quietly ignored"
            }
        """

        Twine(source=VALID_SCHEMA_TWINE).validate_configuration_values(source=configuration_valid_with_extra_field)


if __name__ == "__main__":
    unittest.main()



import os

from twined import Twine, exceptions
from .base import BaseTestCase


class TestManifestStrands(BaseTestCase):
    """Testing operation of the Twine class for validation of data using strands which require manifests"""

    VALID_MANIFEST_STRAND = """
        {
            "configuration_manifest": {
                "datasets": {
                    "configuration_files_data": {
                        "purpose": "A dataset containing files used in configuration"
                    }
                }
            },
            "input_manifest": {
                "datasets": {
                    "met_mast_data": {
                        "purpose": "A dataset containing meteorological mast data"
                    },
                    "scada_data": {
                        "purpose": "A dataset containing scada data"
                    }
                }
            },
            "output_manifest": {
                "datasets": {
                    "output_files_data": {
                        "purpose": "A dataset containing output results"
                    }
                }
            }
        }
    """

    def test_missing_manifest_files(self):
        """Ensures that if you try to read values from missing files, the right exceptions get raised"""
        twine = Twine(source=self.VALID_MANIFEST_STRAND)
        file = os.path.join(self.path, "not_a_file.json")

        with self.assertRaises(exceptions.ConfigurationManifestFileNotFound):
            twine.validate_configuration_manifest(source=file)

        with self.assertRaises(exceptions.InputManifestFileNotFound):
            twine.validate_input_manifest(source=file)

        with self.assertRaises(exceptions.OutputManifestFileNotFound):
            twine.validate_output_manifest(source=file)

    def test_error_raised_if_datasets_are_missing_from_manifest(self):
        """Test that an error is raised if a dataset is missing from a manifest."""
        twine = """
            {
                "input_manifest": {
                    "datasets": {
                        "cat": {
                            "purpose": "blah"
                        },
                        "dog": {
                            "purpose": "blah"
                        }
                    }
                }
            }
        """

        input_manifest = {
            "id": "30d2c75c-a7b9-4f16-8627-9c8d5cc04bf4",
            "datasets": {"my-dataset": "gs://my-bucket/my_dataset", "dog": "gs://dog-house/dog"},
        }

        twine = Twine(source=twine)

        with self.assertRaises(exceptions.InvalidManifestContents) as context:
            twine.validate_input_manifest(source=input_manifest)

        self.assertEqual(
            context.exception.message,
            "A dataset named 'cat' is expected in the input_manifest but is missing.",
        )

    def test_missing_optional_datasets_do_not_raise_error(self):
        """Test that optional datasets specified in the twine missing from the manifest don't raise an error."""
        twine = """
            {
                "input_manifest": {
                    "datasets": {
                        "cat": {
                            "purpose": "blah",
                            "optional": true
                        },
                        "dog": {
                            "purpose": "blah"
                        }
                    }
                }
            }
        """

        input_manifest = {
            "id": "30d2c75c-a7b9-4f16-8627-9c8d5cc04bf4",
            "datasets": {"dog": "gs://dog-house/dog"},
        }

        Twine(source=twine).validate_input_manifest(source=input_manifest)

    def test_valid_manifest_files(self):
        """Ensures that a manifest file will validate."""
        valid_configuration_manifest = """
            {
                "id": "3ead7669-8162-4f64-8cd5-4abe92509e17",
                "datasets": {
                    "configuration_files_data": {
                        "id": "34ad7669-8162-4f64-8cd5-4abe92509e17",
                        "name": "configuration_files_data",
                        "tags": {},
                        "labels": ["the", "config", "labels"],
                        "files": [
                            {
                                "path": "configuration/datasets/7ead7669/file_1.csv",
                                "tags": {},
                                "labels": [],
                                "timestamp": 0,
                                "id": "abff07bc-7c19-4ed5-be6d-a6546eae8e86"
                            },
                            {
                                "path": "configuration/datasets/7ead7669/file_2.csv",
                                "tags": {},
                                "labels": [],
                                "timestamp": 0,
                                "id": "bbff07bc-7c19-4ed5-be6d-a6546eae8e45"
                            }
                        ]
                    }
                }
            }
        """

        valid_input_manifest = """
            {
                "id": "8ead7669-8162-4f64-8cd5-4abe92509e17",
                "datasets": {
                    "met_mast_data": {
                        "id": "7ead7669-8162-4f64-8cd5-4abe92509e17",
                        "name": "met_mast_data",
                        "tags": {},
                        "labels": ["met", "mast", "wind"],
                        "files": [
                            {
                                "path": "input/datasets/7ead7669/file_1.csv",
                                "tags": {},
                                "labels": [],
                                "timestamp": 0,
                                "id": "abff07bc-7c19-4ed5-be6d-a6546eae8e86"
                            },
                            {
                                "path": "input/datasets/7ead7669/file_2.csv",
                                "tags": {},
                                "labels": [],
                                "timestamp": 0,
                                "id": "bbff07bc-7c19-4ed5-be6d-a6546eae8e45"
                            }
                        ]
                    },
                    "scada_data": "gs://my-bucket/scada-data"
                }
            }
        """

        valid_output_manifest = """
            {
                "id": "2ead7669-8162-4f64-8cd5-4abe92509e17",
                "datasets": {
                    "output_files_data": {
                        "id": "1ead7669-8162-4f64-8cd5-4abe92509e17",
                        "name": "output_files_data",
                        "tags": {},
                        "labels": ["the", "output", "labels"],
                        "files": [
                            {
                                "path": "input/datasets/7ead7669/file_1.csv",
                                "tags": {},
                                "labels": [],
                                "timestamp": 0,
                                "id": "abff07bc-7c19-4ed5-be6d-a6546eae8e86"
                            },
                            {
                                "path": "input/datasets/7ead7669/file_2.csv",
                                "tags": {},
                                "labels": [],
                                "timestamp": 0,
                                "id": "bbff07bc-7c19-4ed5-be6d-a6546eae8e45"
                            }
                        ]
                    }
                }
            }
        """

        twine = Twine(source=self.VALID_MANIFEST_STRAND)
        twine.validate_configuration_manifest(source=valid_configuration_manifest)
        twine.validate_input_manifest(source=valid_input_manifest)
        twine.validate_output_manifest(source=valid_output_manifest)

    # def test_empty_values(self):
    #     """ Ensures that appropriate errors are generated for invalid values
    #     """
    #     twine_file = VALID_SCHEMA_TWINE
    #     twine = Twine(file=twine_file)
    #     values_file = os.path.join(self.path, "configurations", "empty.json")
    #     with self.assertRaises(exceptions.InvalidValuesJson):
    #         twine.validate_configuration(file=values_file)
    #
    # def test_incorrect_values(self):
    #     """ Ensures that appropriate errors are generated for invalid values
    #     """
    #     twine_file = VALID_SCHEMA_TWINE
    #     twine = Twine(file=twine_file)
    #     values_file = os.path.join(self.path, "configurations", "incorrect.json")
    #     with self.assertRaises(exceptions.InvalidValuesContents):
    #         twine.validate_configuration(file=values_file)
    #
    # def test_missing_not_required_values(self):
    #     """ Ensures that appropriate errors are generated for missing values
    #     """
    #     twine_file = VALID_SCHEMA_TWINE
    #     twine = Twine(file=twine_file)
    #     values_file = os.path.join(self.path, "outputs", "missing_not_required.json")
    #     twine.validate_output_values(file=values_file)
    #
    # def test_missing_required_values(self):
    #     """ Ensures that appropriate errors are generated for missing values
    #     """
    #     twine_file = VALID_SCHEMA_TWINE
    #     twine = Twine(file=twine_file)
    #     values_file = os.path.join(self.path, "inputs", "missing_required.json")
    #     with self.assertRaises(exceptions.InvalidValuesContents):
    #         twine.validate_input_values(file=values_file)
    #
    # def test_valid_values_files(self):
    #     """ Ensures that values can be read and validated correctly from files on disk
    #     """
    #     twine_file = VALID_SCHEMA_TWINE
    #     twine = Twine(file=twine_file)
    #     twine.validate_configuration(file=os.path.join(self.path, "configurations", "valid.json"))
    #     twine.validate_input_values(file=os.path.join(self.path, "inputs", "valid.json"))
    #     twine.validate_output_values(file=os.path.join(self.path, "outputs", "valid.json"))
    #
    # def test_valid_values_json(self):
    #     """ Ensures that values can be read and validated correctly from a json string
    #     """
    #     twine_file = VALID_SCHEMA_TWINE
    #     twine = Twine(file=twine_file)
    #     values_file = os.path.join(self.path, "configurations", "valid.json")
    #     with open(values_file, "r", encoding="utf-8") as f:
    #         json_string = f.read()
    #     twine.validate_configuration(json=json_string)
    #
    # def test_valid_with_extra_values(self):
    #     """ Ensures that extra values get ignored
    #     """
    #     twine_file = VALID_SCHEMA_TWINE
    #     twine = Twine(file=twine_file)
    #     values_file = os.path.join(self.path, "configurations", "valid_with_extra.json")
    #     twine.validate_configuration(file=values_file)

    def test_error_raised_if_multiple_datasets_have_same_name(self):
        """Test that an error is raised if the input manifest has more than one dataset with the same name."""
        twine = """
            {
                "input_manifest": {
                    "datasets": {
                        "met_mast_data": {
                            "purpose": "A dataset containing meteorological mast data"
                        }
                    }
                }
            }
        """

        input_manifest = """
            {
                "id": "8ead7669-8162-4f64-8cd5-4abe92509e17",
                "datasets": {
                    "met_mast_data": {
                        "id": "7ead7669-8162-4f64-8cd5-4abe92509e19",
                        "name": "met_mast_data",
                        "tags": {},
                        "labels": [],
                        "files": []
                    },
                    "met_mast_data": {
                        "id": "7ead7669-8162-4f64-8cd5-4abe92509e18",
                        "name": "met_mast_data",
                        "tags": {},
                        "labels": [],
                        "files": []
                    }
                }
            }
        """

        twine = Twine(source=twine)

        with self.assertRaises(KeyError):
            twine.validate_input_manifest(source=input_manifest)



import os

from twined import Twine, exceptions
from .base import BaseTestCase


class TestTwine(BaseTestCase):
    """Testing operation of the Twine class"""

    def test_init_twine_with_filename(self):
        """Ensures that the twine class can be instantiated with a file"""
        Twine(source=os.path.join(self.path, "apps", "simple_app", "twine.json"))

    def test_init_twine_with_json(self):
        """Ensures that a twine can be instantiated with a json string"""
        with open(os.path.join(self.path, "apps", "simple_app", "twine.json"), "r", encoding="utf-8") as f:
            Twine(source=f.read())

    def test_no_twine(self):
        """Tests that the canonical-but-useless case of no twine provided validates empty"""
        Twine()

    def test_incorrect_version_twine(self):
        """Ensures exception is thrown on mismatch between installed and specified versions of twined"""
        incorrect_version_twine = """{"twined_version": "0.0.0"}"""
        with self.assertRaises(exceptions.TwineVersionConflict):
            Twine(source=incorrect_version_twine)

    def test_empty_twine(self):
        """Ensures that an empty twine file can be loaded"""
        with self.assertLogs(level="DEBUG") as log:
            Twine(source="{}")
            self.assertEqual(len(log.output), 3)
            self.assertEqual(len(log.records), 3)
            self.assertIn("Detected source", log.output[0])
            self.assertIn("Validated", log.output[1])

    def test_example_twine(self):
        """Ensures that the example (full) twine can be loaded and validated"""
        Twine(source=os.path.join(self.path, "apps", "example_app", "twine.json"))

    def test_simple_twine(self):
        """Ensures that the simple app schema can be loaded and used to parse some basic config and values data"""
        Twine(source=os.path.join(self.path, "apps", "simple_app", "twine.json"))

    def test_broken_json_twine(self):
        """Ensures that an invalid json file raises an InvalidTwine exception"""
        invalid_json_twine = """
            {
                "children": [
                "configuration_values_schema": {
                    "$schema": "https://json-schema.org/draft/2020-12/schema",
                    "title": "The example configuration form",
                    "description": "The configuration strand of an example twine",
                    "type": "object",
                    "properties": {
                    }
                },
            }
        """

        with self.assertRaises(exceptions.InvalidTwineJson):
            Twine(source=invalid_json_twine)

    def test_available_strands_properties(self):
        """Test that the `available_strands` and `available_manifest_strands` properties work correctly."""
        twine = """
            {
                "configuration_values_schema": {
                    "$schema": "https://json-schema.org/draft/2020-12/schema",
                    "title": "The example configuration form",
                    "description": "The configuration strand of an example twine",
                    "type": "object",
                    "properties": {
                        "n_iterations": {
                            "description": "An example of an integer configuration variable, called 'n_iterations'.",
                            "type": "integer",
                            "minimum": 1,
                            "maximum": 10,
                            "default": 5
                        }
                    }
                },
                "input_values_schema": {
                    "$schema": "https://json-schema.org/draft/2020-12/schema",
                    "title": "Input Values",
                    "description": "The input values strand of an example twine, with a required height value",
                    "type": "object",
                    "properties": {
                        "height": {
                            "description": "An example of an integer value called 'height'",
                            "type": "integer",
                            "minimum": 2
                        }
                    },
                    "required": ["height"]
                },
                "output_values_schema": {
                    "title": "Output Values",
                    "description": "The output values strand of an example twine",
                    "type": "object",
                    "properties": {
                        "width": {
                            "description": "An example of an integer value called 'result'",
                            "type": "integer",
                            "minimum": 2
                        }
                    }
                },
                "output_manifest": {
                    "datasets": {
                        "my-dataset": {}
                    }
                }
            }
        """

        twine = Twine(source=twine)

        self.assertEqual(
            twine.available_strands,
            {"configuration_values", "input_values", "output_values", "output_manifest"},
        )

        self.assertEqual(twine.available_manifest_strands, {"output_manifest"})
